@@ -1,40 +1,45 @@
+import asyncio
+import getpass
+import logging
 import sys
 from contextlib import contextmanager
 from datetime import datetime
 from enum import Enum, auto
 
 import click
-import logging
-import getpass
-
 import tzlocal
-from asn1crypto import pem, cms
-
-from pyhanko.pdf_utils.layout import LayoutError
+from asn1crypto import cms, pem
 from pyhanko_certvalidator import ValidationContext
-from pyhanko.config import (
-    init_validation_context_kwargs, parse_cli_config,
-    CLIConfig, LogConfig, StdLogOutput, parse_logging_config,
-    PKCS11SignatureConfig, PKCS12SignatureConfig, PemDerSignatureConfig
-)
-from pyhanko.pdf_utils import misc
-from pyhanko.pdf_utils.config_utils import ConfigurationError
 
-from pyhanko.sign import signers, validation, fields
-from pyhanko.sign.general import (
-    SigningError, SignatureValidationError, load_certs_from_pemder
+from pyhanko import __version__
+from pyhanko.config import (
+    CLIConfig,
+    LogConfig,
+    PemDerSignatureConfig,
+    PKCS11SignatureConfig,
+    PKCS12SignatureConfig,
+    StdLogOutput,
+    init_validation_context_kwargs,
+    parse_cli_config,
+    parse_logging_config,
 )
-from pyhanko.sign.signers import DEFAULT_SIGNER_KEY_USAGE
-from pyhanko.sign.timestamps import HTTPTimeStamper
+from pyhanko.pdf_utils import crypt, misc
+from pyhanko.pdf_utils.config_utils import ConfigurationError
+from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
+from pyhanko.pdf_utils.layout import LayoutError
 from pyhanko.pdf_utils.reader import PdfFileReader
 from pyhanko.pdf_utils.writer import copy_into_new_writer
-from pyhanko.pdf_utils.incremental_writer import IncrementalPdfFileWriter
-from pyhanko.pdf_utils import crypt
-from pyhanko.sign.validation import (
-    RevocationInfoValidationType
+from pyhanko.sign import fields, signers, validation
+from pyhanko.sign.general import (
+    SignatureValidationError,
+    SigningError,
+    load_certs_from_pemder,
 )
-from pyhanko.stamp import QRStampStyle, text_stamp_file, qr_stamp_file
-from pyhanko import __version__
+from pyhanko.sign.signers import DEFAULT_SIGNER_KEY_USAGE
+from pyhanko.sign.signers.pdf_cms import PdfCMSSignedAttributes
+from pyhanko.sign.timestamps import HTTPTimeStamper
+from pyhanko.sign.validation import RevocationInfoValidationType
+from pyhanko.stamp import QRStampStyle, qr_stamp_file, text_stamp_file
 
 __all__ = ['cli']
 
@@ -156,14 +161,19 @@ def cli(ctx, config, verbose):
         log_config[None] = LogConfig(
             level=logging.DEBUG, output=root_logger_config.output
         )
-    elif 'fontTools.subset' not in log_config:
-        # the fontTools subsetter has a very noisy INFO log, so
-        # set that one to WARNING by default
-        log_config['fontTools.subset'] = LogConfig(
-            level=logging.WARNING,
-            # use the root logger's output settings to populate the default
-            output=log_config[None].output
+    else:
+        # use the root logger's output settings to populate the default
+        log_output = log_config[None].output
+        # Revinfo fetch logs -> filter by default
+        log_config['pyhanko_certvalidator.fetchers'] = LogConfig(
+            level=logging.WARNING, output=log_output
         )
+        if 'fontTools.subset' not in log_config:
+            # the fontTools subsetter has a very noisy INFO log, so
+            # set that one to WARNING by default
+            log_config['fontTools.subset'] = LogConfig(
+                level=logging.WARNING, output=log_output
+            )
 
     logging_setup(log_config)
 
@@ -363,22 +373,23 @@ def _validate_detached(infile, sig_infile, validation_context,
     except ValueError as e:
         raise click.ClickException("Could not parse CMS object") from e
 
-    return validation.validate_detached_cms(
+    validation_coro = validation.async_validate_detached_cms(
         infile, signed_data=content_info['content'],
         signer_validation_context=validation_context,
         key_usage_settings=key_usage_settings
     )
+    return asyncio.run(validation_coro)
 
 
 def _signature_status_str(status_callback, pretty_print, executive_summary):
     try:
         status = status_callback()
         if executive_summary and not pretty_print:
-            return 'VALID' if status.bottom_line else 'INVALID'
+            return ('VALID' if status.bottom_line else 'INVALID', status.bottom_line)
         elif pretty_print:
-            return status.pretty_print_details()
+            return (status.pretty_print_details(), status.bottom_line)
         else:
-            return status.summary()
+            return (status.summary(), status.bottom_line)
     except validation.ValidationInfoReadingError as e:
         msg = (
             'An error occurred while parsing the revocation information '
@@ -386,16 +397,16 @@ def _signature_status_str(status_callback, pretty_print, executive_summary):
         )
         logger.error(msg)
         if pretty_print:
-            return msg
+            return (msg, False)
         else:
-            return 'REVINFO_FAILURE'
+            return ('REVINFO_FAILURE', False)
     except SignatureValidationError as e:
         msg = 'An error occurred while validating this signature: ' + str(e)
         logger.error(msg, exc_info=e)
         if pretty_print:
-            return msg
+            return (msg, False)
         else:
-            return 'INVALID'
+            return ('INVALID', False)
 
 
 # TODO add an option to do LTV, but guess the profile
@@ -468,14 +479,17 @@ def validate_signatures(ctx, infile, executive_summary,
     )
     with pyhanko_exception_manager():
         if detached is not None:
-            status_str = _signature_status_str(
+            (status_str, signature_ok) = _signature_status_str(
                 status_callback=lambda: _validate_detached(
                     infile, detached, ValidationContext(**vc_kwargs),
                     key_usage_settings
                 ),
                 pretty_print=pretty_print, executive_summary=executive_summary
             )
-            print(status_str)
+            if signature_ok:
+                print(status_str)
+            else:
+                raise click.ClickException(status_str)
             return
 
         r = PdfFileReader(infile)
@@ -492,9 +506,10 @@ def validate_signatures(ctx, infile, executive_summary,
                 "validating (for now)"
             )
 
+        all_signatures_ok = True
         for ix, embedded_sig in enumerate(r.embedded_regular_signatures):
             fingerprint: str = embedded_sig.signer_cert.sha256.hex()
-            status_str = _signature_status_str(
+            (status_str, signature_ok) = _signature_status_str(
                 status_callback=lambda: _signature_status(
                     ltv_profile=ltv_profile, force_revinfo=force_revinfo,
                     vc_kwargs=vc_kwargs, key_usage_settings=key_usage_settings,
@@ -513,6 +528,10 @@ def validate_signatures(ctx, infile, executive_summary,
                 print('\n\n' + status_str)
             else:
                 print('%s:%s:%s' % (name, fingerprint, status_str))
+            all_signatures_ok &= signature_ok
+
+        if not all_signatures_ok:
+            raise click.ClickException("Validation failed")
 
 
 @signing.command(name='list', help='list signature fields')
@@ -745,6 +764,14 @@ def get_text_params(ctx):
 
 def detached_sig(signer: signers.Signer, infile_path, outfile,
                  timestamp_url, use_pem):
+    coro = async_detached_sig(
+        signer, infile_path, outfile, timestamp_url, use_pem
+    )
+    return asyncio.run(coro)
+
+
+async def async_detached_sig(signer: signers.Signer, infile_path, outfile,
+                             timestamp_url, use_pem):
 
     with pyhanko_exception_manager():
         if timestamp_url is not None:
@@ -752,13 +779,15 @@ def detached_sig(signer: signers.Signer, infile_path, outfile,
             timestamp = None
         else:
             timestamper = None
-            # in this case, embed the signing time as an unsigned attr
+            # in this case, embed the signing time as a signed attr
             timestamp = datetime.now(tz=tzlocal.get_localzone())
 
         with open(infile_path, 'rb') as inf:
-            signature = signer.sign_general_data(
+            signature = await signer.async_sign_general_data(
                 inf, signers.DEFAULT_MD, timestamper=timestamper,
-                timestamp=timestamp
+                signed_attr_settings=PdfCMSSignedAttributes(
+                    signing_time=timestamp
+                )
             )
 
         output_bytes = signature.dump()
@@ -997,6 +1026,9 @@ def _sign_pkcs11(ctx, signer, infile, outfile, timestamp_url):
               required=False)
 @click.option('--cert-label', help='certificate label', type=str,
               required=False)
+@click.option('--raw-mechanism',
+              help='invoke raw PKCS#11 mechanism',
+              type=bool, is_flag=True, required=False)
 @click.option('--key-label', help='key label', type=str, required=False)
 @click.option('--slot-no', help='specify PKCS#11 slot to use',
               required=False, type=int, default=None)
@@ -1008,7 +1040,8 @@ def _sign_pkcs11(ctx, signer, infile, outfile, timestamp_url):
                    'other options)')
 @click.pass_context
 def addsig_pkcs11(ctx, infile, outfile, lib, token_label,
-                  cert_label, key_label, slot_no, skip_user_pin, p11_setup):
+                  cert_label, key_label, slot_no, skip_user_pin, p11_setup,
+                  raw_mechanism):
     from pyhanko.sign import pkcs11
     timestamp_url = ctx.obj[Ctx.TIMESTAMP_URL]
 
@@ -1033,17 +1066,20 @@ def addsig_pkcs11(ctx, infile, outfile, lib, token_label,
         pkcs11_config = PKCS11SignatureConfig(
             module_path=lib, cert_label=cert_label, key_label=key_label,
             slot_no=slot_no, token_label=token_label,
-            prompt_pin=not skip_user_pin
+            prompt_pin=not skip_user_pin, raw_mechanism=raw_mechanism
         )
 
-    with pkcs11.PKCS11SigningContext(pkcs11_config) as signer:
+    pin = pkcs11_config.user_pin
+    if pin is None and pkcs11_config.prompt_pin:  # pragma: nocover
+        pin = getpass.getpass(prompt='PKCS#11 user PIN: ')
+    with pkcs11.PKCS11SigningContext(pkcs11_config, user_pin=pin) as signer:
         _sign_pkcs11(ctx, signer, infile, outfile, timestamp_url)
 
 
 @click.argument('infile', type=readable_file)
 @click.argument('outfile', type=click.File('wb'))
 @click.option('--lib', help='path to libbeidpkcs11 library file',
-              type=readable_file, required=True)
+              type=readable_file, required=False)
 @click.option('--use-auth-cert', type=bool, show_default=True,
               default=False, required=False, is_flag=True,
               help='use Authentication cert instead')
@@ -1052,6 +1088,15 @@ def addsig_pkcs11(ctx, infile, outfile, lib, token_label,
 @click.pass_context
 def addsig_beid(ctx, infile, outfile, lib, use_auth_cert, slot_no):
     from pyhanko.sign import beid
+    if not lib:
+        cli_config: CLIConfig = ctx.obj.get(Ctx.CLI_CONFIG, None)
+        if cli_config is None or cli_config.beid_module_path is None:
+            raise click.ClickException(
+                "The --lib option is mandatory unless beid-module-path is "
+                "provided in the configuration file."
+            )
+        lib = cli_config.beid_module_path
+
     timestamp_url = ctx.obj[Ctx.TIMESTAMP_URL]
     session = beid.open_beid_session(lib, slot_no=slot_no)
     with session:
