@@ -1,21 +1,19 @@
+import binascii
 import enum
 import logging
-from datetime import timedelta
-from typing import Dict, Optional, Union, List, Iterable
 from dataclasses import dataclass
+from datetime import timedelta
+from typing import Dict, Iterable, List, Optional, Union
 
 import yaml
 from asn1crypto import x509
+from pyhanko_certvalidator import ValidationContext
 
 from pyhanko.pdf_utils import config_utils
-from pyhanko_certvalidator import ValidationContext
-from pyhanko.pdf_utils.config_utils import (
-    check_config_keys, ConfigurationError
-)
+from pyhanko.pdf_utils.config_utils import ConfigurationError, check_config_keys
 from pyhanko.pdf_utils.misc import get_and_apply
-
-from pyhanko.sign import load_certs_from_pemder, SimpleSigner
-from pyhanko.sign.general import KeyUsageConstraints
+from pyhanko.sign import SimpleSigner, load_certs_from_pemder
+from pyhanko.sign.general import KeyUsageConstraints, load_cert_from_pemder
 from pyhanko.sign.signers import DEFAULT_SIGNING_STAMP_STYLE
 from pyhanko.stamp import QRStampStyle, TextStampStyle
 
@@ -64,6 +62,7 @@ class CLIConfig:
     pemder_setups: Dict[str, dict]
     pkcs12_setups: Dict[str, dict]
     pkcs11_setups: Dict[str, dict]
+    beid_module_path: Optional[str]
 
     # TODO graceful error handling for syntax & type issues?
 
@@ -382,19 +381,39 @@ class PKCS11SignatureConfig(config_utils.ConfigurableMixin):
     module_path: str
     """Path to the PKCS#11 module shared object."""
 
-    token_label: str
-    """PKCS#11 token name"""
-
-    cert_label: str
+    cert_label: Optional[str] = None
     """PKCS#11 label of the signer's certificate."""
 
+    cert_id: Optional[bytes] = None
+    """PKCS#11 ID of the signer's certificate."""
+
+    signing_certificate: Optional[x509.Certificate] = None
+    """
+    The signer's certificate. If present, :attr:`cert_id` and
+    :attr:`cert_label` will not be used to obtain the signer's certificate
+    from the PKCS#11 token.
+
+    .. note::
+        This can be useful in case the signer's certificate is not available on
+        the token, or if you would like to present a different certificate than
+        the one provided on the token.
+    """
+
+    token_label: Optional[str] = None
+    """PKCS#11 token name"""
+
     other_certs: List[x509.Certificate] = None
     """Other relevant certificates."""
 
     key_label: Optional[str] = None
     """
-    PKCS#11 label of the signer's private key, if different from
-    :attr:`cert_label`.
+    PKCS#11 label of the signer's private key. Defaults to :attr:`cert_label`
+    if the latter is specified and :attr:`key_id` is not.
+    """
+
+    key_id: Optional[bytes] = None
+    """
+    PKCS#11 key ID.
     """
 
     slot_no: Optional[int] = None
@@ -443,6 +462,14 @@ class PKCS11SignatureConfig(config_utils.ConfigurableMixin):
     Prefer PSS to PKCS#1 v1.5 padding when creating RSA signatures.
     """
 
+    raw_mechanism: bool = False
+    """
+    Invoke the raw variant of the PKCS#11 signing operation.
+
+    .. note::
+        This is currently only supported for ECDSA signatures.
+    """
+
     @classmethod
     def process_entries(cls, config_dict):
         super().process_entries(config_dict)
@@ -451,6 +478,43 @@ class PKCS11SignatureConfig(config_utils.ConfigurableMixin):
             other_certs = (other_certs,)
         config_dict['other_certs'] = list(load_certs_from_pemder(other_certs))
 
+        if 'token_label' not in config_dict and 'slot_no' not in config_dict:
+            raise ConfigurationError(
+                "Either 'slot_no' or 'token_label' must be provided in "
+                "PKCS#11 setup"
+            )
+
+        cert_file = config_dict.get('signing_certificate', None)
+        if cert_file is not None:
+            config_dict['signing_certificate'] \
+                = load_cert_from_pemder(cert_file)
+
+        if 'key_id' in config_dict:
+            config_dict['key_id'] \
+                = _process_pkcs11_id_value(config_dict['key_id'])
+        elif 'key_label' not in config_dict and 'cert_label' not in config_dict:
+            raise ConfigurationError(
+                "Either 'key_id', 'key_label' or 'cert_label' must be provided "
+                "in PKCS#11 setup"
+            )
+
+        if 'cert_id' in config_dict:
+            config_dict['cert_id'] \
+                = _process_pkcs11_id_value(config_dict['cert_id'])
+        elif 'cert_label' not in config_dict \
+                and 'signing_certificate' not in config_dict:
+            raise ConfigurationError(
+                "Either 'cert_id', 'cert_label' or 'signing_certificate' "
+                "must be provided in PKCS#11 setup"
+            )
+
+
+def _process_pkcs11_id_value(x: Union[str, int]):
+    if isinstance(x, int):
+        return bytes([x])
+    else:
+        return binascii.unhexlify(x)
+
 
 DEFAULT_VALIDATION_CONTEXT = DEFAULT_STAMP_STYLE = 'default'
 DEFAULT_TIME_TOLERANCE = 10
@@ -460,9 +524,12 @@ STAMP_STYLE_TYPES = {
 }
 
 
-def parse_cli_config(yaml_str):
+def parse_cli_config(yaml_str) -> CLIConfig:
     config_dict = yaml.safe_load(yaml_str) or {}
+    return CLIConfig(**process_config_dict(config_dict))
 
+
+def process_config_dict(config_dict: dict) -> dict:
     # validation context config
     vcs = {DEFAULT_VALIDATION_CONTEXT: {}}
     try:
@@ -485,9 +552,11 @@ def parse_cli_config(yaml_str):
     log_config_spec = config_dict.get('logging', {})
     log_config = parse_logging_config(log_config_spec)
 
+    # TODO type check!
     pkcs11_setups = config_dict.get('pkcs11-setups', {})
     pkcs12_setups = config_dict.get('pkcs12-setups', {})
     pemder_setups = config_dict.get('pemder-setups', {})
+    beid_module_path = config_dict.get('beid-module-path', None)
 
     # some misc settings
     default_vc = config_dict.get(
@@ -506,10 +575,11 @@ def parse_cli_config(yaml_str):
 
     time_tolerance = timedelta(seconds=time_tolerance_seconds)
     retroactive_revinfo = bool(config_dict.get('retroactive-revinfo', False))
-    return CLIConfig(
+    return dict(
         validation_contexts=vcs, default_validation_context=default_vc,
         time_tolerance=time_tolerance, retroactive_revinfo=retroactive_revinfo,
         stamp_styles=stamp_configs, default_stamp_style=default_stamp_style,
         log_config=log_config, pkcs11_setups=pkcs11_setups,
-        pkcs12_setups=pkcs12_setups, pemder_setups=pemder_setups
+        pkcs12_setups=pkcs12_setups, pemder_setups=pemder_setups,
+        beid_module_path=beid_module_path
     )
