@@ -5,25 +5,28 @@ The internals were imported from PyPDF2, with modifications.
 See :ref:`here <pypdf2-license>` for the original license
 of the PyPDF2 project.
 """
+import binascii
+import codecs
+import decimal
+import logging
 import os
 import re
-import binascii
+from dataclasses import dataclass, field
 from datetime import datetime, timedelta, timezone
 from io import BytesIO
-from typing import Iterator, Tuple, Optional, Union, Callable, Any
-from dataclasses import dataclass, field
+from typing import Any, Callable, Iterator, Optional, Tuple, Union
 
-from .misc import (
-    read_non_whitespace, skip_over_comment, read_until_regex,
-    is_regular_character
-)
-from .misc import (
-    PdfStreamError, PdfReadError, IndirectObjectExpected, PdfWriteError
-)
-import logging
 from . import filters
-import decimal
-import codecs
+from .misc import (
+    IndirectObjectExpected,
+    PdfReadError,
+    PdfStreamError,
+    PdfWriteError,
+    is_regular_character,
+    read_non_whitespace,
+    read_until_regex,
+    skip_over_comment,
+)
 
 __all__ = [
     'Dereferenceable', 'Reference', 'TrailerReference',
@@ -849,6 +852,7 @@ def _decode_name(name_bytes: bytes) -> 'NameObject':
             )
     except StopIteration:
         pass
+    name_bytes = result.getvalue()
     # NOTE: we assume UTF-8, but the PDF spec actually doesn't prescribe
     #  a character encoding for names, they're just byte sequences.
     #  This doesn't matter in 99.99% of cases (since names are not supposed
@@ -857,7 +861,18 @@ def _decode_name(name_bytes: bytes) -> 'NameObject':
     #  massive non-obvious API breakage (since NameObject inherits from 'str' as
     #  in PyPDF2), i.e. the correctness benefit is vastly outweighed by the
     #  risks (for now)
-    return NameObject(result.getvalue().decode('utf8'))
+    encodings_to_try = ('utf8', 'latin1')
+    # latin1 should never trigger decoding errors, since Python's implementation
+    # maps even unassigned values to corresponding unicode codepoints
+    name_str = None
+    for enc in encodings_to_try:
+        try:
+            name_str = name_bytes.decode(enc)
+            break
+        except ValueError:
+            pass
+    assert name_str is not None
+    return NameObject(name_str)
 
 
 class NameObject(str, PdfObject):
@@ -1531,7 +1546,7 @@ ASN_DT_FORMAT = "D:%Y%m%d%H%M%S"
 def pdf_date(dt: datetime) -> TextStringObject:
     """
     Convert a datetime object into a PDF string.
-    This funciton supports both timezone-aware and naive datetimes.
+    This function supports both timezone-aware and naive datetime objects.
 
     :param dt:
         The datetime object to convert.
@@ -1542,8 +1557,8 @@ def pdf_date(dt: datetime) -> TextStringObject:
     base_dt = dt.strftime(ASN_DT_FORMAT)
     utc_offset_string = ''
     if dt.tzinfo is not None:
-        # compute UTC off set string
-        tz_seconds = dt.utcoffset().seconds
+        # compute UTC offset string
+        tz_seconds = dt.utcoffset().total_seconds()
         if not tz_seconds:
             utc_offset_string = 'Z'
         else:
