@@ -10,23 +10,23 @@ This comes at a cost, and future iterations of this module may offer more
 flexibility in terms of the level of detail with which file size is scrutinised.
 """
 
-import struct
+import logging
 import os
 import re
+import struct
 from collections import defaultdict
 from io import BytesIO
 from itertools import chain
-from typing import Set, List, Optional, Union, Tuple
+from typing import List, Optional, Set, Tuple, Union
 
 from . import generic, misc
-from .misc import PdfReadError
 from .crypt import (
-    SecurityHandler, StandardSecurityHandler,
-    EnvelopeKeyDecrypter, PubKeySecurityHandler,
+    EnvelopeKeyDecrypter,
+    PubKeySecurityHandler,
+    SecurityHandler,
+    StandardSecurityHandler,
 )
-
-import logging
-
+from .misc import PdfReadError
 from .rw_common import PdfHandler
 
 logger = logging.getLogger(__name__)
@@ -1540,15 +1540,21 @@ class HistoricalResolver(PdfHandler):
 
         pages_ref = self.root.raw_get('/Pages')
         page_tree_nodes = set()
-        page_tree_nodes.update(
-            _collect_page_tree_refs(pages_obj=pages_ref.get_object())
-        )
+        for ref in _collect_page_tree_refs(pages_obj=pages_ref.get_object()):
+            if ref in page_tree_nodes:
+                raise misc.PdfReadError(
+                    "Circular reference in page tree in mapping stage"
+                )
+            page_tree_nodes.add(ref)
         struct_tree_nodes = set()
         try:
             struct_tree_root_ref = self.root.raw_get('/StructTreeRoot')
-            struct_tree_nodes.update(
-                _collect_struct_tree_refs(struct_tree_root_ref.get_object())
-            )
+            for ref in _collect_struct_tree_refs(struct_tree_root_ref.get_object()):
+                if ref in struct_tree_nodes:
+                    raise misc.PdfReadError(
+                        "Circular reference in structure tree in mapping stage"
+                    )
+                struct_tree_nodes.add(ref)
         except KeyError:
             pass
         _compute_paths_to_refs(
