@@ -5,28 +5,30 @@ Contains code from the PyPDF2 project; see :ref:`here <pypdf2-license>`
 for the original license.
 """
 
+import enum
 import os
 import struct
-import enum
 from dataclasses import dataclass
 from io import BytesIO
-from typing import List, Union, Optional, Tuple, Iterable
+from typing import Iterable, List, Optional, Tuple, Union
 
 from asn1crypto import x509
 
-from pyhanko.pdf_utils import generic, content
+from pyhanko import __version__
+from pyhanko.pdf_utils import content, generic
 from pyhanko.pdf_utils.crypt import (
-    SecurityHandler, StandardSecurityHandler,
     PubKeySecurityHandler,
+    SecurityHandler,
+    StandardSecurityHandler,
 )
 from pyhanko.pdf_utils.generic import pdf_name, pdf_string
 from pyhanko.pdf_utils.misc import (
-    peek, PdfReadError, instance_test,
+    PdfReadError,
     PdfWriteError,
+    instance_test,
+    peek,
 )
 from pyhanko.pdf_utils.rw_common import PdfHandler
-from pyhanko import __version__
-
 
 __all__ = [
     'ObjectStream', 'BasePdfFileWriter',
@@ -56,15 +58,17 @@ class ObjectStream:
         legacy XRef table cannot contain object streams either.
         See § 7.5.7 in ISO 32000-1 for further details.
 
-    .. warning::
-        The usefulness of object streams is somewhat stymied by the fact that
-        PDF stream objects cannot be embedded into object streams for
-        syntactical reasons.
+    .. danger::
+        Use :meth:`.BasePdfFileWriter.prepare_object_stream` to create instances
+        of object streams. The `__init__` function is internal API.
+
     """
 
-    def __init__(self, compress=True):
+    def __init__(self, writer: 'BasePdfFileWriter', compress=True):
         self._obj_refs = {}
         self.compress = compress
+        self.writer = writer
+        self._ref = None
 
     def add_object(self, idnum: int, obj: generic.PdfObject):
         """
@@ -88,6 +92,23 @@ class ObjectStream:
             )  # pragma: nocover
         self._obj_refs[idnum] = obj
 
+    def register_and_emit(self):
+        """
+        Internal method to flush an object stream as part of the file
+        writing process.
+        """
+        stream_ref = self._ref
+        objects = self._obj_refs
+        if objects and stream_ref is None:
+            # first, register the object stream object
+            #  (will get written later)
+            stream_ref = self._ref \
+                = self.writer.add_object(self.as_pdf_object())
+        # loop over all objects in the stream, and prepare
+        # the data to put in the XRef table
+        for ix, (idnum, obj) in enumerate(objects.items()):
+            yield idnum, (stream_ref.idnum, ix)
+
     def as_pdf_object(self) -> generic.StreamObject:
         """
         Render the object stream to a PDF stream object
@@ -248,6 +269,10 @@ def _contiguous_xref_chunks(position_dict):
     previous_idnum = None
     current_chunk = []
 
+    if not position_dict.keys():
+        # return immediately, there are no objects
+        return
+
     # iterate over keys in object ID order
     key_iter = sorted(position_dict.keys(), key=lambda t: t[1])
     (_, first_idnum), key_iter = peek(key_iter)
@@ -288,7 +313,12 @@ def _write_xref_table(stream, position_dict):
             entry = "%010d %05d n \n" % (position, generation)
             stream.write(entry.encode('ascii'))
 
-    first_idnum, subsection = next(subsections)
+    try:
+        first_idnum, subsection = next(subsections)
+    except StopIteration:
+        # no updates, just write '0 0' and be done with it
+        stream.write(b'0 0\n')
+        return xref_location
     # TODO support deleting objects
     # case distinction: in contrast with the above we have to ensure that
     # everything is written in one chunk when *not* doing incremental updates.
@@ -726,7 +756,7 @@ class BasePdfFileWriter(PdfHandler):
             raise PdfWriteError(
                 'Object streams require Xref streams to be enabled.'
             )
-        stream = ObjectStream(compress=compress)
+        stream = ObjectStream(self, compress=compress)
         self.object_streams.append(stream)
         return stream
 
@@ -740,13 +770,8 @@ class BasePdfFileWriter(PdfHandler):
     def _write_objects(self, stream, object_position_dict):
         # deal with objects in object streams first
         for obj_stream in self.object_streams:
-            # first, register the object stream object
-            #  (will get written later)
-            stream_ref = self.add_object(obj_stream.as_pdf_object())
-            # loop over all objects in the stream, and prepare
-            # the data to put in the XRef table
-            for ix, (idnum, obj) in enumerate(obj_stream._obj_refs.items()):
-                object_position_dict[(0, idnum)] = (stream_ref.idnum, ix)
+            for idnum, pos_record in obj_stream.register_and_emit():
+                object_position_dict[(0, idnum)] = pos_record
 
         for ix in sorted(self.objects.keys()):
             generation, idnum = ix
@@ -1400,7 +1425,8 @@ def copy_into_new_writer(input_handler: PdfHandler) -> PdfFileWriter:
         New :class:`.PdfFileWriter` containing all objects from the input
         handler.
     """
-    w = PdfFileWriter(init_page_tree=False)
+    # TODO try to be more clever with object streams
+    w = PdfFileWriter(init_page_tree=False, stream_xrefs=False)
     input_root_ref = input_handler.root_ref
     output_root_ref = w.root_ref
     # call _import_object in such a way that we translate the input handler's
