@@ -53,24 +53,34 @@ putting together custom diff rules.
 
 """
 
-import re
 import logging
+import re
 from collections import defaultdict
 from dataclasses import dataclass
 from enum import unique
 from io import BytesIO
 from typing import (
-    Iterable, Optional, Set, Tuple, Generator, TypeVar, Dict,
-    List, Callable, Union, Iterator
+    Callable,
+    Dict,
+    Generator,
+    Iterable,
+    Iterator,
+    List,
+    Optional,
+    Set,
+    Tuple,
+    TypeVar,
+    Union,
 )
 
-from pyhanko.pdf_utils.generic import Reference, PdfObject
+from pyhanko.pdf_utils import generic, misc
+from pyhanko.pdf_utils.generic import PdfObject, Reference
 from pyhanko.pdf_utils.misc import OrderedEnum
 from pyhanko.pdf_utils.reader import (
-    HistoricalResolver, PdfFileReader,
+    HistoricalResolver,
+    PdfFileReader,
     RawPdfPath,
 )
-from pyhanko.pdf_utils import generic, misc
 from pyhanko.sign.fields import FieldMDPSpec, MDPPerm
 
 __all__ = [
@@ -345,20 +355,49 @@ class DocInfoRule(WhitelistRule):
         )
 
 
+def _assert_stream_refs(der_obj_type, arr, err_cls, is_vri):
+    arr = arr.get_object()
+    all_stream_arr = isinstance(arr, generic.ArrayObject) and all(
+        isinstance(obj, generic.IndirectObject)
+        and isinstance(obj.get_object(), generic.StreamObject)
+        for obj in arr
+    )
+    if not all_stream_arr:
+        raise err_cls(
+            f"Expected contents of '{der_obj_type}' in "
+            f"{'VRI' if is_vri else 'DSS'} to be an array of stream references."
+        )
+
+
 def _validate_dss_substructure(old: HistoricalResolver, new: HistoricalResolver,
-                               new_dict, der_stream_keys, is_vri):
+                               old_dict, new_dict, der_stream_keys, is_vri,
+                               path: RawPdfPath):
     for der_obj_type in der_stream_keys:
+        as_update = ReferenceUpdate.curry_ref(paths_checked=path + der_obj_type)
         try:
             value = new_dict.raw_get(der_obj_type)
         except KeyError:
             continue
-        if not isinstance(value.get_object(), generic.ArrayObject):
-            raise SuspiciousModification(
-                f"Expected array at {'VRI' if is_vri else 'DSS'} "
-                f"key {der_obj_type}."
-            )
+        _assert_stream_refs(der_obj_type, value, SuspiciousModification, is_vri)
+        if isinstance(value, generic.IndirectObject):
+            new_ref = value.reference
+            try:
+                old_value = old_dict.raw_get(der_obj_type)
+                if isinstance(old_value, generic.IndirectObject):
+                    yield from map(
+                        as_update,
+                        _safe_whitelist(old, old_value.reference, new_ref)
+                    )
+                _assert_stream_refs(
+                    der_obj_type, old_value, misc.PdfReadError, is_vri
+                )
+                # We don't enforce the contents of the new array vs. the old one
+                # deleting info is allowed by PAdES, and this check can get
+                # pretty expensive.
+            except KeyError:
+                pass
 
-        yield from map(ReferenceUpdate, new.collect_dependencies(
+        yield from map(as_update, new.collect_dependencies(
             value, since_revision=old.revision + 1
         ))
 
@@ -370,8 +409,10 @@ class DSSCompareRule(WhitelistRule):
     This rule will validate the structure of the DSS quite rigidly, and
     will raise :class:`.SuspiciousModification` whenever it encounters
     structural problems with the DSS.
-    Similarly, modifications that remove items from the DSS also count as
-    suspicious.
+    Similarly, modifications that remove structural items from the DSS
+    also count as suspicious. However, merely removing individual OCSP
+    responses, CRLs or certificates when they become irrelevant is permitted.
+    This is also allowed by PAdES.
     """
 
     def apply(self, old: HistoricalResolver, new: HistoricalResolver)\
@@ -404,7 +445,8 @@ class DSSCompareRule(WhitelistRule):
             )
 
         yield from _validate_dss_substructure(
-            old, new, new_dss, dss_der_stream_keys, is_vri=False
+            old, new, old_dss, new_dss, dss_der_stream_keys, is_vri=False,
+            path=RawPdfPath('/Root', '/DSS')
         )
 
         # check that the /VRI dictionary still contains all old keys, unchanged.
@@ -470,7 +512,9 @@ class DSSCompareRule(WhitelistRule):
                     f"{new_vri_value_keys - vri_expected_keys}."
                 )
             yield from _validate_dss_substructure(
-                old, new, new_vri_dict, vri_der_stream_keys, is_vri=True
+                old, new, generic.DictionaryObject(),
+                new_vri_dict, vri_der_stream_keys, is_vri=True,
+                path=RawPdfPath('/Root', '/DSS', '/VRI', key)
             )
 
             # /TS is also a DER stream
@@ -883,7 +927,8 @@ class SigFieldCreationRule(FieldMDPRule):
             _walk_page_tree_annots(
                 old_page_root, new_page_root,
                 field_ref_reverse, context.old,
-                valid_when_locked=True
+                valid_when_locked=True,
+                refs_seen=set()
             )
         )
 
@@ -1361,8 +1406,8 @@ class MetadataUpdateRule(WhitelistRule):
                 "/Metadata should be a reference to a stream object"
             )
 
-        from xml.sax.handler import ContentHandler
         from xml.sax import make_parser
+        from xml.sax.handler import ContentHandler
 
         parser = make_parser()
         parser.setContentHandler(ContentHandler())
@@ -1610,7 +1655,7 @@ def _extract_annots_from_page(page, exc):
 
 def _walk_page_tree_annots(old_page_root, new_page_root,
                            field_name_dict, old: HistoricalResolver,
-                           valid_when_locked):
+                           valid_when_locked, refs_seen):
     def get_kids(page_root, exc):
         try:
             return _arr_to_refs(page_root['/Kids'], exc)
@@ -1627,6 +1672,10 @@ def _walk_page_tree_annots(old_page_root, new_page_root,
             "Unexpected change to page tree structure."
         )
     for new_kid_ref, old_kid_ref in zip(new_kids, old_kids):
+        if old_kid_ref in refs_seen:
+            raise misc.PdfReadError(
+                "Circular reference in page tree during annotation analysis"
+            )
         new_kid = new_kid_ref.get_object()
         old_kid = old_kid_ref.get_object()
         try:
@@ -1636,7 +1685,7 @@ def _walk_page_tree_annots(old_page_root, new_page_root,
         if node_type == '/Pages':
             yield from _walk_page_tree_annots(
                 old_kid, new_kid, field_name_dict, old,
-                valid_when_locked
+                valid_when_locked, refs_seen | {old_kid_ref},
             )
         elif node_type == '/Page':
             try:
