@@ -6,37 +6,42 @@ CMS is defined in :rfc:`5652`. To parse CMS messages, pyHanko relies heavily on
 `asn1crypto <https://github.com/wbond/asn1crypto>`_.
 """
 
+import hashlib
 import logging
 from dataclasses import dataclass
-from typing import ClassVar, Set, Optional, Tuple, Union
+from typing import ClassVar, Optional, Set, Tuple, Union
 
-import hashlib
-
-from asn1crypto import x509, cms, tsp, algos, pem, keys
+from asn1crypto import algos, cms, keys, pem, tsp, x509
 
 # noinspection PyProtectedMember
 from cryptography.exceptions import InvalidSignature
+from cryptography.hazmat.primitives import hashes, serialization
+from cryptography.hazmat.primitives.asymmetric import padding
+from cryptography.hazmat.primitives.asymmetric.dsa import DSAPublicKey
 from cryptography.hazmat.primitives.asymmetric.ec import (
-    EllipticCurvePublicKey, ECDSA
+    ECDSA,
+    EllipticCurvePublicKey,
 )
 from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
 from cryptography.hazmat.primitives.asymmetric.utils import Prehashed
-
-from pyhanko.pdf_utils.config_utils import ConfigurableMixin, \
-    process_bit_string_flags, process_oids
-from pyhanko_certvalidator.path import ValidationPath
-
 from pyhanko_certvalidator import CertificateValidator
-from pyhanko_certvalidator.registry import (
-    CertificateStore, SimpleCertificateStore
-)
 from pyhanko_certvalidator.errors import (
-    RevokedError, PathValidationError, InvalidCertificateError,
-    PathBuildingError
+    InvalidCertificateError,
+    PathBuildingError,
+    PathValidationError,
+    RevokedError,
+)
+from pyhanko_certvalidator.path import ValidationPath
+from pyhanko_certvalidator.registry import (
+    CertificateStore,
+    SimpleCertificateStore,
 )
-from cryptography.hazmat.primitives import serialization, hashes
-from cryptography.hazmat.primitives.asymmetric import padding
 
+from pyhanko.pdf_utils.config_utils import (
+    ConfigurableMixin,
+    process_bit_string_flags,
+    process_oids,
+)
 
 __all__ = [
     'SignatureStatus', 'simple_cms_attribute',
@@ -344,8 +349,8 @@ class SignatureStatus:
             return 'INVALID'
 
     @classmethod
-    def validate_cert_usage(cls, validator: CertificateValidator,
-                            key_usage_settings: KeyUsageConstraints = None):
+    async def validate_cert_usage(cls, validator: CertificateValidator,
+                                  key_usage_settings: KeyUsageConstraints = None):
         key_usage_settings = key_usage_settings or KeyUsageConstraints()
         key_usage_settings = KeyUsageConstraints(
             key_usage=(
@@ -365,7 +370,7 @@ class SignatureStatus:
         try:
             # validate usage without going through pyhanko_certvalidator
             key_usage_settings.validate(cert)
-            path = validator.validate_usage(key_usage=set())
+            path = await validator.async_validate_usage(key_usage=set())
             trusted = True
         except InvalidCertificateError as e:
             # TODO accumulate these somewhere
@@ -825,6 +830,9 @@ def _validate_raw(signature: bytes, signed_data: bytes, cert: x509.Certificate,
             prehashed=prehashed
         )
         pub_key.verify(signature, signed_data, pss_padding, hash_algo)
+    elif sig_algo == 'dsa':
+        assert isinstance(pub_key, DSAPublicKey)
+        pub_key.verify(signature, signed_data, verify_md)
     elif sig_algo == 'ecdsa':
         assert isinstance(pub_key, EllipticCurvePublicKey)
         pub_key.verify(signature, signed_data, ECDSA(verify_md))
